from typing import Type, get_type_hints, Any, Dict, List
from pydantic import BaseModel
import inspect
from dataclasses import dataclass

@dataclass
class ProtoField:
    name: str
    type: str
    number: int
    required: bool = True
    description: str = ""

@dataclass
class ProtoMessage:
    name: str
    fields: List[ProtoField]
    package: str = "wabee.tools"
    
class ProtoSchemaGenerator:
    TYPE_MAPPING = {
        int: "int64",
        str: "string",
        float: "double",
        bool: "bool",
        dict: "bytes",  # For complex nested structures
        list: "repeated bytes",  # For arrays
    }
    
    @classmethod
    def generate_from_pydantic(
        cls,
        model: Type[BaseModel],
        message_name: str
    ) -> ProtoMessage:
        fields = []
        for i, (name, field) in enumerate(model.model_fields.items(), start=1):
            proto_type = cls.TYPE_MAPPING.get(field.annotation.__class__, "bytes")
            fields.append(ProtoField(
                name=name,
                type=proto_type,
                number=i,
                required=field.is_required(),
                description=field.description or ""
            ))
        return ProtoMessage(message_name, fields)

    @classmethod
    def generate_from_function(
        cls,
        func: Any,
        message_name: str
    ) -> ProtoMessage:
        hints = get_type_hints(func)
        sig = inspect.signature(func)
        
        fields = []
        for i, (name, param) in enumerate(sig.parameters.items(), start=1):
            if name == 'self' or name == 'cls':
                continue
                
            param_type = hints.get(name, Any)
            proto_type = cls.TYPE_MAPPING.get(param_type, "bytes")
            required = param.default == inspect.Parameter.empty
            
            fields.append(ProtoField(
                name=name,
                type=proto_type,
                number=i,
                required=required
            ))
            
        return ProtoMessage(message_name, fields)

    @classmethod
    def get_tool_schema(cls, tool: Any) -> Dict[str, Any]:
        """Get schema information for a tool"""
        if hasattr(tool, 'args_schema'):
            schema = tool.args_schema.model_json_schema()
        else:
            hints = get_type_hints(tool)
            sig = inspect.signature(tool)
            schema = {
                "properties": {
                    name: {"type": str(typ)}
                    for name, typ in hints.items()
                    if name != "return"
                },
                "required": [
                    name for name, param in sig.parameters.items()
                    if param.default == inspect.Parameter.empty
                ]
            }
        return schema




from dataclasses import dataclass
from enum import Enum
from typing import Optional

class ToolErrorType(Enum):
    RETRYABLE = "retryable"
    PERMANENT = "permanent"
    INVALID_INPUT = "invalid_input"
    INTERNAL_ERROR = "internal_error"
    VALIDATION_ERROR = "validation_error"
    EXECUTION_ERROR = "execution_error"

@dataclass
class ToolError:
    type: ToolErrorType
    message: str
    original_error: Optional[Exception] = None



import os
import re
import shutil
from typing import Literal, Optional, Type

def to_camel_case(name: str) -> str:
    """
    Convert a string to CamelCase.
    Example: 'my_cool_tool' -> 'MyCoolTool'
    """
    # First sanitize the name to ensure valid Python identifier
    sanitized = sanitize_name(name)
    # Split by underscore and capitalize each word
    return ''.join(word.capitalize() for word in sanitized.split('_'))

def sanitize_name(name: str) -> str:
    """
    Sanitize the tool name to be a valid Python identifier.
    
    - Converts to lowercase
    - Replaces spaces and invalid chars with underscores
    - Ensures it starts with a letter
    - Removes consecutive underscores
    """
    # Convert to lowercase and replace spaces/invalid chars with underscore
    name = re.sub(r'[^a-zA-Z0-9_]', '_', name.lower())
    
    # Ensure starts with letter
    if not name[0].isalpha():
        name = 'tool_' + name
        
    # Remove consecutive underscores
    name = re.sub(r'_+', '_', name)
    
    # Remove trailing underscore
    name = name.rstrip('_')
    
    return name

class CreateToolService:
    TOOL_TYPES = Literal["simple", "complete"]
    
    def __init__(self):
        self.templates_dir = os.path.join(os.path.dirname(__file__), "templates")
        
    def create_tool(
        self, 
        name: str, 
        tool_type: TOOL_TYPES,
        description: str,
        version: str,
        tool_language: Literal["python", "javascript"] = "python",
        generate_js: bool = True
    ) -> None:
        """Create a new tool project with the given name and type."""
        # Sanitize the name
        sanitized_name = sanitize_name(name)
        # Get CamelCase version for class names
        camel_case_name = to_camel_case(name)
        
        # Create project directory
        os.makedirs(sanitized_name, exist_ok=True)
        
        if tool_language == "python":
            # Create Python files
            self._create_python_files(sanitized_name, camel_case_name, tool_type, description, version)
            
            if generate_js:
                # Create JavaScript client
                self._create_js_files(sanitized_name, camel_case_name, description, version)
        else:  # javascript
            # Create only JavaScript files
            self._create_js_files(sanitized_name, camel_case_name, description, version)

    def _create_python_files(
        self,
        sanitized_name: str,
        camel_case_name: str,
        tool_type: TOOL_TYPES,
        description: str,
        version: str
    ) -> None:
        """Create Python-specific files for the tool."""
        # Create tool file
        tool_file = os.path.join(sanitized_name, f"{sanitized_name}_tool.py")
        with open(tool_file, "w") as f:
            template_args = {
                "name": sanitized_name,
                "class_name": camel_case_name,
                "description": description
            }
            if tool_type == "simple":
                f.write(self._get_simple_tool_template(sanitized_name, camel_case_name))
            else:
                f.write(self._get_complete_tool_template(sanitized_name, camel_case_name))
        
        # Create toolspec.yaml
        toolspec_file = os.path.join(sanitized_name, "toolspec.yaml")
        with open(toolspec_file, "w") as f:
            f.write(self._get_toolspec_template(camel_case_name, description, version))

        # Create requirements.txt
        requirements_file = os.path.join(sanitized_name, "requirements.txt")
        with open(requirements_file, "w") as f:
            f.write(self._get_requirements_template(sanitized_name, description, version))
                
        # Create server.py
        self._create_server_file(sanitized_name)

    def _create_js_files(
        self,
        sanitized_name: str,
        camel_case_name: str,
        description: str,
        version: str
    ) -> None:
        """Create JavaScript/TypeScript files for the tool."""
        # Create package.json
        package_file = os.path.join(sanitized_name, "package.json")
        with open(package_file, "w") as f:
            f.write(self._get_js_package_template(sanitized_name, description, version))
        
        # Create TypeScript client
        src_dir = os.path.join(sanitized_name, "src")
        os.makedirs(src_dir, exist_ok=True)
        client_file = os.path.join(src_dir, "index.ts")
        with open(client_file, "w") as f:
            f.write(self._get_ts_client_template(sanitized_name, camel_case_name))
        
        # Copy the server.js template
        server_template_path = os.path.join(os.path.dirname(__file__), "templates", "js_server.js")
        shutil.copy2(server_template_path, os.path.join(sanitized_name, "server.js"))
        
        # Create tsconfig.json
        tsconfig_file = os.path.join(sanitized_name, "tsconfig.json")
        with open(tsconfig_file, "w") as f:
            f.write(self._get_tsconfig_template())
        
    def _get_simple_tool_template(self, snake_name: str, class_name: str) -> str:
        return f'''from typing import Optional
from pydantic import BaseModel
from wabee.tools.simple_tool import simple_tool
from wabee.tools.tool_error import ToolError

class {class_name}Input(BaseModel):
    message: str

@simple_tool(
    name="{class_name}",
    description="A simple tool that processes the input message",
    schema={class_name}Input
)
async def {snake_name.lower()}_tool(input_data: {class_name}Input) -> str:
    """
    A simple tool that processes the input message.
    
    Args:
        input_data: The input data containing the message to process
        
    Returns:
        The processed message
    """
    return f"Processed: {{input_data.message}}"
'''

    def _get_complete_tool_template(self, snake_name: str, class_name: str) -> str:
        return f'''from typing import Optional, Type
from pydantic import BaseModel
from wabee.tools.base_tool import BaseTool
from wabee.tools.tool_error import ToolError, ToolErrorType

class {class_name}Input(BaseModel):
    message: str

class {class_name}Tool(BaseTool):
    args_schema: Type[BaseModel] = {class_name}Input

    def __init__(self, **kwargs):
        """Initialize the tool with configuration."""
        super().__init__(
            name="{class_name}",
            description="A tool that processes the input message",
            **kwargs
        )

    async def execute(self, input_data: {class_name}Input) -> tuple[Optional[str], Optional[ToolError]]:
        """
        Execute the tool's main functionality.
        
        Args:
            input_data: The validated input data
            
        Returns:
            A tuple of (result, error) where one will be None
        """
        try:
            result = f"Processed: {{input_data.message}}"
            return result, None
        except Exception as e:
            return None, ToolError(
                type=ToolErrorType.EXECUTION_ERROR,
                message=str(e),
                original_error=e
            )

    @classmethod
    def create(cls, **kwargs) -> "{class_name}Tool":
        """Factory method to create an instance of this tool."""
        return cls(**kwargs)
'''

    def _get_toolspec_template(self, name: str, description: str, version: str) -> str:
        return f'''tool:
  name: {name}
  description: {description}
  version: {version}
  entrypoint: {name.lower()}_tool.py
'''

    def _get_requirements_template(self, name: str, description: str, version: str) -> str:
        return '''wabee>=0.2.6
pydantic>=2.0.0
'''

    def _get_js_package_template(self, name: str, description: str, version: str) -> str:
        return f'''{{
  "name": "{name}",
  "version": "{version}",
  "description": "{description}",
  "main": "dist/index.js",
  "types": "dist/index.d.ts",
  "scripts": {{
    "build": "npx tsc",
    "start": "node server.js",
    "test": "jest"
  }},
  "dependencies": {{
    "@wabee_ai/sdk": "^0.1.3",
    "@grpc/grpc-js": "^1.8.0",
    "@grpc/proto-loader": "^0.7.0",
    "zod": "^3.21.0"
  }},
  "devDependencies": {{
    "@types/node": "^18.0.0",
    "typescript": "^4.9.0",
    "jest": "^29.0.0",
    "ts-jest": "^29.0.0",
    "@types/jest": "^29.0.0"
  }}
}}'''

    def _get_ts_client_template(self, name: str, class_name: str) -> str:
        return f'''import {{ z }} from 'zod';
import {{ simpleTool, ToolOptions }} from '@wabee_ai/sdk';

export const {class_name}Schema = z.object({{
    message: z.string().describe("Message to be displayed")
}});

export type {class_name}Input = z.infer<typeof {class_name}Schema>;

export function create{class_name}Tool(options?: ToolOptions) {{
    return simpleTool(
        '{name}',
        {class_name}Schema,
        options
    );
}}
'''

    def _get_tsconfig_template(self) -> str:
        return '''{
  "compilerOptions": {
    "target": "es2020",
    "module": "commonjs",
    "declaration": true,
    "outDir": "./dist",
    "strict": true,
    "esModuleInterop": true,
    "skipLibCheck": true,
    "forceConsistentCasingInFileNames": true
  },
  "include": ["src"],
  "exclude": ["node_modules", "dist", "**/*.test.ts"]
}'''

    def _create_server_file(self, name: str) -> None:
        """Create the server.py file in the tool directory."""
        template_path = os.path.join(os.path.dirname(__file__), "templates", "python_server.py")
        shutil.copy2(template_path, os.path.join(name, "server.py"))



from functools import wraps
from typing import Callable, Type, Optional, Any, Union, TypeVar, Awaitable
from typing_extensions import ParamSpec
from pydantic import BaseModel, create_model, ConfigDict

from wabee.tools.base_tool import BaseTool
from wabee.tools.tool_error import ToolError, ToolErrorType

T = TypeVar('T')
P = ParamSpec('P')

def simple_tool(
    name: Optional[str] = None,
    description: Optional[str] = None,
    schema: Optional[Type[BaseModel]] = None,
    **schema_fields: Any
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[tuple[Optional[T], Optional[ToolError]]]]]:
    """
    A decorator that transforms a simple async function into a BaseTool-compatible interface.
    
    Can be used in these ways:
    1. With inline schema fields: @simple_tool(name="Add", description="Adds numbers", x=int, y=int)
    2. With a predefined schema: @simple_tool(name="Add", description="Adds numbers", schema=MySchema)
    3. Without any schema: @simple_tool(name="Add", description="Adds numbers")
    4. With automatic name/description: @simple_tool()
    
    Args:
        name: Optional name for the tool (defaults to function name)
        description: Optional description (defaults to function docstring)
        schema: Optional predefined Pydantic model for input validation
        **schema_fields: Field definitions to create an ad-hoc Pydantic model
        
    Returns:
        A decorator that wraps the function in a BaseTool-compatible interface
        
    Example:
        @simple_tool(x=int, y=int)
        async def add_numbers(input_data):
            return input_data.x + input_data.y
            
        result, error = await add_numbers(x=5, y=3)
    """
    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[tuple[Optional[T], Optional[ToolError]]]]:
        # Create a schema on the fly if fields are provided but no schema
        if schema is None and schema_fields:
            # Convert field definitions to proper Pydantic field annotations
            annotated_fields = {
                field_name: (field_type, ...)  # ... means required
                for field_name, field_type in schema_fields.items()
            }
            model_name = f"{func.__name__.title()}Input"
            dynamic_schema = create_model(
                model_name,
                __config__=ConfigDict(arbitrary_types_allowed=True),
                __module__=func.__module__,
                __base__=None,
                **annotated_fields
            )
        else:
            dynamic_schema = schema

        @wraps(func)
        async def wrapped_tool(*args: P.args, **kwargs: P.kwargs) -> tuple[Union[T, None], Optional[ToolError]]:
            # Get tool name and description
            tool_name = name or func.__name__
            tool_description = description or func.__doc__ or ""
            
            # Create an anonymous class inheriting from BaseTool
            class FunctionalTool(BaseTool):
                args_schema = dynamic_schema

                def __init__(self):
                    super().__init__(name=tool_name, description=tool_description)

                async def execute(self, input_data: Any) -> tuple[Union[T, None], Optional[ToolError]]:
                    try:
                        # Validate and convert input to Pydantic model if schema exists
                        if dynamic_schema:
                            if isinstance(input_data, dict):
                                validated_input = dynamic_schema(**input_data)
                            elif isinstance(input_data, dynamic_schema):
                                validated_input = input_data
                            else:
                                raise ValueError(f"Input must be dict or {dynamic_schema.__name__}")
                            result = await func(validated_input)
                        else:
                            # For functions without schema, validate against type hints
                            try:
                                # Create a dynamic model from function annotations
                                hints = func.__annotations__
                                if hints:
                                    fields = {
                                        name: (typ, ...) for name, typ in hints.items()
                                        if name != 'return'
                                    }
                                    model_name = f"{func.__name__}RuntimeSchema"
                                    runtime_schema = create_model(
                                        model_name,
                                        __config__=ConfigDict(arbitrary_types_allowed=True),
                                        __module__=func.__module__,
                                        __base__=None,
                                        **fields
                                    )
                                    # Validate kwargs against the runtime schema
                                    validated_kwargs = runtime_schema(**kwargs).dict()
                                    result = await func(**validated_kwargs)
                                else:
                                    # When no schema and no type hints, pass args/kwargs directly 
                                    if args:
                                        result = await func(*args)
                                    else:
                                        result = await func(**kwargs)
                            except (ValueError, TypeError) as e:
                                return None, ToolError(
                                    type=ToolErrorType.INVALID_INPUT,
                                    message=str(e),
                                    original_error=e
                                )
                        return result, None
                    except ValueError as e:
                        # Business logic errors raised by the function
                        return None, ToolError(
                            type=ToolErrorType.EXECUTION_ERROR,
                            message=str(e),
                            original_error=e
                        )
                    except AttributeError as e:
                        # Schema validation/access errors
                        return None, ToolError(
                            type=ToolErrorType.INVALID_INPUT,
                            message=str(e),
                            original_error=e
                        )
                    except Exception as e:
                        # Unexpected errors
                        return None, ToolError(
                            type=ToolErrorType.INTERNAL_ERROR,
                            message=str(e),
                            original_error=e
                        )

            # Create and call the tool instance
            tool = FunctionalTool()
            return await tool(kwargs if dynamic_schema else args[0] if args else kwargs)

        return wrapped_tool

    return decorator



import json
import grpc
from typing import Any, Optional, Dict, Union

from wabee.rpc.protos import tool_service_pb2
from wabee.rpc.protos import tool_service_pb2_grpc

class ToolServiceClient:
    def __init__(
        self,
        host: str = "localhost",
        port: int = 50051,
        use_json: bool = True
    ):
        self.host = host
        self.port = port
        self.use_json = use_json
        self.channel = grpc.aio.insecure_channel(f"{self.host}:{self.port}")
        self.stub = tool_service_pb2_grpc.ToolServiceStub(self.channel)
        
    async def __aenter__(self):
        return self
        
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()
        self.channel = grpc.aio.insecure_channel(f"{self.host}:{self.port}")
        self.stub = tool_service_pb2_grpc.ToolServiceStub(self.channel)
        self.use_json = self.use_json

    async def get_tool_schema(
        self,
        tool_name: str
    ) -> Dict[str, Any]:
        """Get the schema for a specific tool"""
        request = tool_service_pb2.GetToolSchemaRequest(
            tool_name=tool_name
        )
        
        try:
            response = await self.stub.GetToolSchema(request)
            return {
                "tool_name": response.tool_name,
                "fields": [
                    {
                        "name": field.name,
                        "type": field.type,
                        "required": field.required,
                        "description": field.description
                    }
                    for field in response.fields
                ]
            }
        except grpc.RpcError as e:
            return {"error": str(e)}

    async def execute(
        self,
        tool_name: str,
        input_data: Dict[str, Any]
    ) -> tuple[Optional[Any], Optional[Dict]]:
        """Execute a tool with the given input data"""
        try:
            request = tool_service_pb2.ExecuteRequest(
                tool_name=tool_name
            )
            
            if self.use_json:
                request.json_data = json.dumps(input_data)
            else:
                request.proto_data = json.dumps(input_data).encode()
            
            response = await self.stub.Execute(request)
            
            if response.HasField('error'):
                return None, {
                    'type': response.error.type,
                    'message': response.error.message
                }
            
            if response.HasField('json_result'):
                return json.loads(response.json_result), None
            else:
                return json.loads(response.proto_result.decode()), None
            
        except grpc.RpcError as e:
            return None, {
                'type': 'RPC_ERROR',
                'message': str(e)
            }

    async def close(self):
        await self.channel.close()

async def execute_tool(
    tool_name: str,
    input_data: Dict[str, Any],
    host: str = "localhost",
    port: int = 50051
) -> Union[Any, Dict[str, str]]:
    """Convenience function for one-off tool execution"""
    async with ToolServiceClient(host, port) as client:
        result, error = await client.execute(tool_name, input_data)
        if error:
            return error
        return result



import asyncio
import os
import logging
from pathlib import Path
from wabee.rpc.server import serve
from wabee.rpc.loader import ToolLoader

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)

logger = logging.getLogger(__name__)

def main():
    try:
        port = int(os.environ.get('WABEE_GRPC_PORT', '50051'))
        tool = ToolLoader.load_from_spec(Path("toolspec.yaml"))
        logger.info(f"Starting gRPC server on port {port}")
        asyncio.run(serve({tool.tool_name: tool}, port=port))
    except Exception as e:
        logger.error(f"Failed to start server: {e}")
        raise

if __name__ == '__main__':
    main()
